package dev.fitko.fitconnect.client.strategies;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.PublicServiceType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.AttachmentCreationException;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.SchemaNotFoundException;
import dev.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.model.AttachmentPayload;
import dev.fitko.fitconnect.client.model.DataPayload;
import dev.fitko.fitconnect.client.model.SubmissionPayload;
import dev.fitko.fitconnect.client.util.SubmissionUtil;
import org.apache.tika.Tika;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URLConnection;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSubmissionToAnnounce;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.getSchemaUriForMimeType;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.getSubmissionSchemasFromDestination;

public class SendNewSubmissionStrategy implements SubmitStrategy {

    private static final Logger LOGGER = LoggerFactory.getLogger(SendNewSubmissionStrategy.class);

    private final Sender sender;
    private final String metadataSchema;
    private final Environment environment;

    public SendNewSubmissionStrategy(final Sender sender, final String metadataSchema, final Environment environment) {
        this.sender = sender;
        this.metadataSchema = metadataSchema;
        this.environment = environment;
    }

    @Override
    public Optional<SubmitSubmission> send(final SubmissionPayload submissionPayload) {

        final UUID destinationId = submissionPayload.getDestinationId();
        final ServiceType serviceType = submissionPayload.getServiceType();
        final DataPayload userDataPayload = submissionPayload.getData();
        final List<AttachmentPayload> attachments = submissionPayload.getAttachments();

        try {

            /** Get encryption key for destination **/
            final Destination destination = sender.getDestination(destinationId);
            final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId, destination.getEncryptionKid());
            final ValidationResult validationResult = sender.validatePublicKey(encryptionKey);
            if (validationResult.hasError() && !environment.isSilentKeyValidation()) {
                LOGGER.error("Invalid public key", validationResult.getError());
                return Optional.empty();
            }

            /** Create new submission and announce attachments **/
            final List<AttachmentPayload> encryptedAttachments = encryptAndHashAttachments(encryptionKey, attachments);
            final CreateSubmission newSubmission = buildSubmissionToAnnounce(destinationId, serviceType, encryptedAttachments);
            final SubmissionForPickup announcedSubmission = sender.createSubmission(newSubmission);

            final UUID submissionId = announcedSubmission.getSubmissionId();
            uploadAttachments(encryptedAttachments, submissionId);

            /** Create attachment metadata including hashes **/
            final List<Attachment> attachmentMetadata = SubmissionUtil.toAttachmentMetadata(encryptedAttachments);

            /** Build encrypted data payload from user data**/
            LOGGER.info("Adding data payload with mime-type {} to submission", userDataPayload.getMimeType());
            final DataPayload encryptedDataPayload = encryptDataPayload(userDataPayload, destination, encryptionKey);

            /** Set encrypted metadata with data payload and attachments **/
            final Data dataToSend = SubmissionUtil.buildData(encryptedDataPayload);
            final PublicServiceType publicServiceType = SubmissionUtil.buildPublicServiceType(serviceType);
            final Metadata metadata = SubmissionUtil.buildMetadata(attachmentMetadata, dataToSend, publicServiceType);

            final ValidationResult validatedMetadata = sender.validateMetadata(metadata, metadataSchema);
            if (validatedMetadata.hasError()) {
                LOGGER.error("Metadata does not match schema", validatedMetadata.getError());
                return Optional.empty();
            }

            /** Prepare submit submission with announced submission id **/
            final SubmitSubmission submission = new SubmitSubmission();
            submission.setSubmissionId(submissionId);
            submission.setEncryptedData(encryptedDataPayload.getEncryptedData());
            submission.setEncryptedMetadata(sender.encryptObject(encryptionKey, metadata));

            /** Submit submission **/
            sender.sendSubmission(submission);
            LOGGER.info("SUCCESSFULLY HANDED IN SUBMISSION !");
            return Optional.of(submission);

        } catch (final EncryptionException e) {
            LOGGER.error("Encrypting submission failed", e);
        } catch (final RestApiException e) {
            LOGGER.error("Sending submission failed", e);
        } catch (final SchemaNotFoundException e) {
            LOGGER.error("Required schema to send valid submission not found", e);
        } catch (final SubmissionNotCreatedException e) {
            LOGGER.error("Failed to announce new submission", e);
        } catch (final KeyNotRetrievedException e) {
            LOGGER.error("Getting encryption key for destination {} failed", destinationId, e);
        } catch (final AttachmentCreationException e) {
            LOGGER.error("Reading file failed. Attachment will not be created.", e);
        }
        return Optional.empty();
    }

    private DataPayload encryptDataPayload(final DataPayload dataPayload, final Destination destination, final RSAKey encryptionKey) {
        final String hashedData = sender.createHash(dataPayload.getRawData());
        final String encryptedData = sender.encryptBytes(encryptionKey, dataPayload.getRawData());
        final URI schemaUri = getSchemaUri(dataPayload, destination);
        return dataPayload.withSchemaUri(schemaUri)
                .withEncryptedData(encryptedData)
                .withHashedData(hashedData);
    }

    private URI getSchemaUri(final DataPayload dataPayload, final Destination destination) {
        final List<SubmissionSchema> submissionSchemas = getSubmissionSchemasFromDestination(destination);
        final String supportedSchemas = submissionSchemas.stream().map(s -> s.getMimeType().toString()).collect(Collectors.joining(" | "));
        final Optional<URI> schemaUriForMimeType = getSchemaUriForMimeType(submissionSchemas, dataPayload.getMimeType());
        if (schemaUriForMimeType.isEmpty()) {
            LOGGER.error("Destination supports the mime-type '{}'. Mime-type {} is not allowed, please check the destination", supportedSchemas, dataPayload.getMimeType());
            throw new SchemaNotFoundException("Schema for mime-type " + dataPayload.getMimeType() + " not found");
        }
        return schemaUriForMimeType.get();
    }

    private void uploadAttachments(final List<AttachmentPayload> attachmentPayloads, final UUID submissionId) {
        if (attachmentPayloads.isEmpty()) {
            LOGGER.info("No attachments to upload");
        } else {
            LOGGER.info("Uploading {} attachment(s)", attachmentPayloads.size());
        }
        for (final AttachmentPayload payload : attachmentPayloads) {
            sender.uploadAttachment(submissionId, payload.getAttachmentId(), payload.getEncryptedData());
        }
    }

    private List<AttachmentPayload> encryptAndHashAttachments(final RSAKey encryptionKey, final List<AttachmentPayload> attachments) {
        return attachments.stream()
                .filter(Objects::nonNull)
                .map(this::readRawData)
                .map(this::hashBytes)
                .map(payload -> encryptBytes(encryptionKey, payload))
                .map(this::setMimeType)
                .collect(Collectors.toList());
    }

    private AttachmentPayload readRawData(final AttachmentPayload attachmentPayload) {
        final File file = attachmentPayload.getFile();
        try {
            final byte[] rawData = Files.readAllBytes(Paths.get(file.getPath()));
            return attachmentPayload.withRawData(rawData);
        } catch (final IOException e) {
            throw new AttachmentCreationException("Attachment '" + file.getAbsolutePath() + "' could not be read ", e);
        }
    }

    private AttachmentPayload encryptBytes(final RSAKey encryptionKey, final AttachmentPayload attachmentPayload) {
        final String encryptedAttachment = sender.encryptBytes(encryptionKey, attachmentPayload.getRawData());
        return attachmentPayload.withEncryptedData(encryptedAttachment);
    }

    private AttachmentPayload hashBytes(final AttachmentPayload attachmentPayload) {
        final String hashedBytes = sender.createHash(attachmentPayload.getRawData());
        return attachmentPayload.withHashedData(hashedBytes);
    }

    private AttachmentPayload setMimeType(final AttachmentPayload attachmentPayload) {
        final File file = attachmentPayload.getFile();
        String mimeType;
        try {
            final Tika mimeTypeGuesser = new Tika();
            mimeType = mimeTypeGuesser.detect(file);
        } catch (final IOException e) {
            mimeType = URLConnection.guessContentTypeFromName(file.getName());
        }
        LOGGER.info("Detected attachment mime-type {}", mimeType);
        return attachmentPayload.withMimeType(mimeType);
    }
}
